package macondo.mak;

import java.util.Hashtable;
import java.io.*;
import java.net.Socket;

import macondo.util.*;

/**
 *	Loads agent class bytes from a given source.
 *
 *	@author		Andrea Giovannini
 *	@version	beta
 */

public class AgentClassLoader extends ClassLoader {

	private Hashtable	classes = new Hashtable();

	private	String		host;
	private int			port;
	private boolean		save;
	private	String		path;
	
	
	
	// Constructors ------------------------------------------------------
	
	/**
	 *	Creates an <code>AgentClassLoader</code> to load classes
	 *	from a place at the specified host and port.
	 *
	 *	@param	pname	the code base name.
	 *	@param	save	<code>true</code> if classes must be saved,
	 *					<code>false</code> otherwise.
	 *	@param	path	file system path where class files must be saved.
	 *	@exception	IOException		if any error with the connection occurs.
	 */
	public AgentClassLoader(PlaceName pname, boolean save, String path)
			throws IOException {
		this.host = pname.getHost();
		this.port = pname.getPort();
		this.save = save;
		this.path = path;		
	}
	
	// Public Methods ----------------------------------------------------

	/**
	 *	Loads a class with the specified name. This method is called
	 *	from external clients and the class is resolved before it is
	 *	returned.
	 *
	 *	@param	className	the name of the class to load.
	 *	@return				the requested <code>Class</code>, or
	 *						<code>null</code> if it was not found.
	 *	@exception	ClassNotFoundException	if the class loader can't
	 *						find a definition for the class.
	 */
	public Class loadClass(String class_name) throws ClassNotFoundException {
		return loadClass(class_name,true);
	}
	
	/**
	 *	Loads a class with the specified name, resolving it if required.
	 *	This method is called from <code>loadClass(java.lang.String)</code>
	 *	and internally during class resolving.
	 *
	 *	@param	className	the name of the class to load.
	 *	@param	resolve		<code>true</code> if the class must be resolved.
	 *	@return				the requested <code>Class</code>, or
	 *						<code>null</code> if it was not found.
	 *	@exception	ClassNotFoundException	if the class loader can't
	 *						find a definition for the class.	
	 */
	public synchronized Class loadClass(String class_name, boolean resolve)
			throws ClassNotFoundException {
		Class	result;
		byte[]	class_bytes;
		
		// check local cache of classes
		result = (Class)classes.get(class_name);
		
		if (result != null) {
			// return cached result
			return result;
		}
		
		// check with the primordial class loader
		try {
			result = super.findSystemClass(class_name);
			// return local class found in CLASSPATH
			return result;
		} catch (ClassNotFoundException e) {
			// util.exc(e);
		}
		
		if ( !verifyClassName(class_name) ) {
			throw new ClassNotFoundException();
		}
		
		// load class from agent source place
		class_bytes = loadClassBytes(class_name);
		if (class_bytes == null) {
			throw new ClassNotFoundException();
		}
		
		if ( save ) {
			try {
				saveClass(class_name, class_bytes);
			} catch (IOException ioe) {
				// this is a big problem... what to do?
				util.exc(ioe);
			}
		}
		
		// parse class file
		result = defineClass(class_name, class_bytes, 0, class_bytes.length);
		if (result == null) {
			throw new ClassFormatError();
		}
		
		// resolve class if required
		if (resolve) resolveClass(result);
		classes.put(class_name, result);
		return result;
	}
	
	
	// Protected Methods -------------------------------------------------
	
	/**
	 *	Loads class bytes from the specified source.
	 *
	 *	@param	className	the name of the class to load.
	 *	@return				a byte array holding the class file.
	 */
	protected byte[] loadClassBytes(String class_name) {
		Socket				socket = null;
		DataInputStream		input = null;
		ObjectOutputStream	output = null;
		byte[]				buffer = null;
		
		try {
			// open connection with the agent class source
			socket = new Socket(host, port);
			
			input = new DataInputStream(socket.getInputStream());
			output = new ObjectOutputStream(socket.getOutputStream());
		
			// build and send request
			Message req = new Message( Message.CLASS_REQUEST, new ClassRequestInfo(class_name) );
			output.writeObject(req);
			output.flush();
			
			// getting class file length
			int length = input.readInt();
			int available = length;
			
			// now getting class file
			buffer = new byte[length];
			while ( available > 0 ) {
				available -= input.read(buffer, (length - available), available);
			}
		} catch (Exception e) {
			util.exc(e);
		} finally {
			try {
				input.close();
				output.close();
				socket.close();
			} catch (IOException ioe) {}
		}
		return buffer;
	}
	
	// Private Methods ---------------------------------------------------
	
	/**
	 *	Verify that the given class name is not in a forbidden package.
	 *
	 *	@param	className	the class name to be verified.
	 *	@return				true if the class loader can load the class,
	 *						false otherwise.
	 */
	private final boolean verifyClassName(String class_name) {
		if ( class_name.startsWith("java.") || class_name.startsWith("sun.") ) {
			return false;
		}
		
		// Everything is OK
		return true;
	}
	
	/**
	 *	Save a class to stable storage.
	 *
	 *	@param	className	the name of the class to store.
	 *	@param	classBytes	the class bytes to save in a file.
	 *	@exception	IOException		if any I/O error occurs writing
	 *								the file.
	 */
	private final void saveClass(String class_name, byte[] class_bytes)
			throws IOException {
		String	class_file_path;
		
		if (path.endsWith("/")) {
			class_file_path = path + class_name + ".class";
		} else {
			class_file_path = path + "/" + class_name + ".class";
		}
		
		OutputStream os = null;
		
		try {
			os = new FileOutputStream(class_file_path);
			os.write(class_bytes);
		} catch (IOException ioe) {
			// code can't be saved
			throw ioe;
		} finally {
			if ( os != null ) {
				try {
					os.close();
				} catch (IOException ioe1) {}
			}
		}
	}
	
}