package macondo.mak;

/**
 *	A <code>Place</code> has the following functions:
 *	- transport agents across the network;
 *	- execute agents;
 *	- fetch agents code;
 *	- store agents in stable storage;
 *
 *	A <code>Place</code> can be executed as a stand alone server
 *	or can be extended by the programmer.
 *
 *	@author		Andrea Giovannini
 *	@version	beta
 */

import java.util.*;
import java.io.*;
import java.net.*;

import macondo.util.*;

public class Place implements Runnable {

	// IP port of the place
	private int port = 8000;

	// Host name of the place
	private String host = null;

	// Application dependent name
	private String name;

	// flags
	
	// Starts this place in verbose mode
	private static boolean verbose = false;
	// Load agents specified in macondo.agents
	private static boolean load = false;

	private static String	class_path = System.getProperty("java.class.path"),
						file_separator = System.getProperty("file.separator");

		
	// Vector containing the agents executing here
	private Hashtable		agents = null;
	private Vector			stored = null;
	// System properties
	private Properties		properties = null;

	private StoreManager	sm = null;

	private MigrationResult	mresult;

	// Constructors ------------------------------------------------------
	
	public Place() {
		agents = new Hashtable();
		stored = new Vector();
		properties = new Properties();
		mresult = new MigrationResult();
		
		// set local host address
		setLocalAddress();
		
		// load properties from file macondo.properties
		loadProperties();
		
		// init database used to store agents
		initDB();
	}
	
	public Place(String name) {
		super();
		this.name = name;
	}


	// Public Methods ----------------------------------------------------
	
	/**
	 *	Main entry point when <code>Place</code> is used as a
	 *	standalone application.
	 *
	 *	@param	args	command line arguments
	 */
	public static void main(String[] args) {
		// build a new place
		Place p = new Place();
		
		// parse command line switches
		p.parseCmdLine(args);
		
		// start main thread
		p.startPlace();
		
		if ( load ) p.loadAgents();
	}

	/**
	 *	Returns a system property.
	 *
	 *	@return		a system property.
	 */
	public String getProperty(String key) {
		return properties.getProperty(key);
	}

	// Store utilities

	public final void flushAgent(Agent agent, Object key) throws IOException {
		sm.insertObject(key, agent);
	}
	
	public final void saveAgent(Agent agent, Object key) throws IOException {
		sm.insertObject(key, agent);
		// stop agent
		AgentThread at = (AgentThread)agents.remove(agent.getClass().getName());
		at.stop();
		stored.addElement(agent.getClass().getName());
	}

	public final void saveAgent(String agent_name) throws IOException {
		AgentThread at = (AgentThread)agents.remove(agent_name);
		at.stop();
		Agent agent = at.getAgent();
		sm.insertObject(agent_name, agent);
		stored.addElement(agent.getClass().getName());
	}
	
	public final Agent restoreAgent(Object key) throws IOException, ClassNotFoundException {
		Agent	agent = (Agent)sm.getObject(key);
		stored.removeElement(agent.getClass().getName());
		/*
		AgentThread at = new AgentThread(agent, AgentThread.RESTORE);
		at.start();
		*/
		return agent;
	}
	
	public void startPlace() {
		new Thread(this).start();
	}
	
	/**
	 *	Place main loop: accepts requests and serve each one in
	 *	a separate thread.
	 */
	public final void run() {
		ServerSocket acceptSocket = null;
		try {
			acceptSocket = new ServerSocket(port);
			if (verbose) {
				System.out.println("Running place at port " + port);
			}
			// main loop
			while (true) {
				handleConnection( acceptSocket.accept() );
			}
		} catch (IOException ioe) {
			util.exc(ioe);
		} finally {
			if ( acceptSocket != null ) {
				try {
					acceptSocket.close();
				} catch (IOException e) {}
			}
		}
	}

	/**
	 *	Sets the host name of this <code>Place</code>.
	 *
	 *	@param	host	the host name of this <code>Place</code>.
	 */
	public final void setHostName(String host) {
		this.host = host;
	}
	
	/**
	 *	Sets the port number of this <code>Place</code>.
	 *
	 *	@param	port	the port number.
	 */
	public void setPort(int port) {
		this.port = port;
	}
	
	/**
	 *	Sets the application dependent name of this <code>Place</code>.
	 *
	 *	@param	name	the application dependant name
	 *					of this <code>Place</code>.
	 */
	public final void setName(String name) {
		this.name = name;
	}
	
	/**
	 */
	public final PlaceName getPlaceName() {
		return new PlaceName(host,port);
	}
	

	/**
	 *	Launchs an agent given its file name. This is a useful for
	 *	agents specified at the command line.
	 *
	 *	@param		agent_name		the file name of the agent to launch.
	 *	@exception	Exception		thrown if agent can't be launched.
	 */
	public void launchAgent(String agent_name) throws Exception {
		Class agent_class = Class.forName(agent_name);
		Agent agent = (Agent)(agent_class.newInstance());
		launchAgent(agent);
	}
	
	/**
	 *	Launchs the given agent.
	 *
	 *	@param		agent_name		the file name of the agent to launch.
	 *	@exception	Exception		thrown if agent can't be launched.
	 */
	public void launchAgent(Agent agent) throws Exception {
		agent.setPlace(this);
		agent.setSource(getPlaceName());
		Thread at = new AgentThread(agent, Agent.INIT);
		at.start();
		agents.put(agent.getClass().getName(), at);
	}
	
	// Protected Methods -------------------------------------------------

	/**
	 *	Handles a connection which requests a service. A separate thread
	 *	is spawned to manage the connection.
	 *
	 *	@param	socket		the connection socket.
	 */	
	protected void handleConnection(Socket socket) {
		try {
			new RequestFilter(socket);
		} catch (IOException e) {
			e.printStackTrace();
		}
	}

	// Package Methods ---------------------------------------------------
	
	/**
	 *	Moves the agent to the requested destination.
	 *
	 *	@param	anAgent		the agent to move
	 *	@param	dest		destination place name
	 *	@exception	TransportException			never occurs. One of the
	 *		exception is thrown instead.
	 *	@exception	UnknownDestinationException	occurs if the destination
	 *		place is unknown.
	 *	@exception	TransportFailedException	occurs if migration
	 *		fails for any other reason.
	 */
	final void move(Agent anAgent, PlaceName dest)
			throws TransportException {
		
		Thread t = new MigrationThread(anAgent, dest);
		t.start();
		try {
			t.join();
		} catch (InterruptedException ie) {}

		// block waiting for migration result
		if ( !mresult.success() ) {
			throw mresult.getException();
		} else { // not very clean :)
			Thread.currentThread().stop();
		}
	}

	// Private Methods ---------------------------------------------------

	/**
	 *	Start the execution of an agent.
	 *
	 *	@param	agent	the agent to start.
	 *	@param	mode	the mode used.
	 */
	private void startAgent(Agent agent, int mode) {
		agent.setPlace(this);
		agent.setSource(getPlaceName());
		AgentThread at = new AgentThread(agent, mode);
		at.start();
		agents.put(agent.getClass().getName(), at);
	}

	private void initDB() {
		String dbpath = properties.getProperty("DBPATH");
		/*if (dbpath == null) {
			dbpath = "";
		}*/
		sm = new StoreManager( dbpath );
	}

	/**
	 *	Get local host name.
	 */
	private void setLocalAddress() {
		InetAddress	localAddress = null;
		
		try {
			localAddress = InetAddress.getLocalHost();
			host = localAddress.getHostName();
		} catch (UnknownHostException e) {
			// unable to get local host address
			host = "localhost";
		}
	}

	/**
	 *	Parse command line switches.
	 *
	 *	@param	args	command line arguments
	 */
	private void parseCmdLine(String args[]) {
		for(int i=0; i < args.length; i++) {
			if (args[i].equals("-h") || args[i].equals("--help")) {
				System.out.println("Usage: java Place [-p port] [-v] [-l] [-h[elp]]");
				System.exit(0);
			} else if (args[i].equals("-p")) {
				try {
					port = Integer.parseInt(args[++i]);
				} catch (NumberFormatException e) {
					System.out.println("Bad port number, default used.");
					port = 8000;
				}
			} else if (args[i].equals("-v")) {
				verbose = true;
			} else if (args[i].equals("-l")) {
				load = true;
			} else if (args[i].startsWith("-")) {
				System.out.println("Unknown option.");
				System.exit(-1);
			}
		}
	}

	/**
	 *	Load system properties from initialization file.
	 */
	private void loadProperties() {
		InputStream	is = null;
		
		try {
			is = new FileInputStream("macondo.properties");
		} catch (FileNotFoundException fnfe) {
			//	macondo.properties doesn't exist in user home directory,
			//	trying with current directory
			try {
				is = new FileInputStream(System.getProperty("user.home") + "/macondo.properties");
			} catch (FileNotFoundException fnfe1) {
				util.err("Can't find initialization file. Exiting.");
				System.exit(0);
			}
		}

		// file found, now load properties
		try {
			properties.load(is);
		} catch (IOException e) {
			util.exc(e);
		} finally {
			try {
				is.close();
			} catch (IOException e1) {}
		}
		
	}
	
	/**
	 *	Loads agents requested in a initialization file.
	 */
	private void loadAgents() {
		// wait 5 sec. and load predefined agents
		try { Thread.sleep(5000); } catch (InterruptedException e) {}
		
		String	tmp;
		Vector	initial_guests = new Vector();

		try {
			RandomAccessFile is = new RandomAccessFile("macondo.agents","r");

			while ( (tmp = is.readLine()) != null ) {
				initial_guests.addElement(tmp);
			}

		} catch (Exception e) {
			util.exc(e);
		}

		Enumeration enum = initial_guests.elements();
		String		agentName;
		Class		agentClass;
		Agent		anAgent;

		while (enum.hasMoreElements()) {
			// get agent name and ...
			agentName = (String)enum.nextElement();

			// start agent in the current place
			try {
				agentClass = Class.forName(agentName);
				anAgent	= (Agent)(agentClass.newInstance());
				anAgent.setPlace(this);
				anAgent.setSource( new PlaceName(host,port) );
				AgentThread at = new AgentThread(anAgent, Agent.INIT);
				agents.put(agentName, at);
				at.start();
			} catch (Exception e) {
				util.exc(e);
			}
		}
	}
	
	
	
	// Inner Classes -----------------------------------------------------
	
	/**
	 *	This is an inner class used to handle in a separate thread
	 *	connections from the net.
	 */
	class RequestFilter implements Runnable {
		private Socket						socket;
		private SmartObjectInputStream		input;
		private OutputStream				output;

		
		public RequestFilter(Socket socket) throws IOException {
			this.socket = socket;
			this.input = new SmartObjectInputStream( socket.getInputStream() );
			this.output = socket.getOutputStream();
			new Thread(this).start();
		}
		
		public void run() {
			try {
				Message		request = (Message)input.readObject();
				int			type = request.getType();
				if ( type == Message.AGENT_ENTER_REQUEST ) {
					handleAgentEnter( (AgentEnterInfo)request.getInfo() );
				} else if ( type == Message.CLASS_REQUEST ) {
					handleClassRequest( (ClassRequestInfo)request.getInfo() );
				}
			} catch (Exception e) {
				util.exc(e);
			} finally {
				try {
					input.close();
					output.close();
					socket.close();
				} catch (IOException e) {
					util.exc(e);
				}
			}
		}
	
		private void handleAgentEnter(AgentEnterInfo req) throws IOException {
			PlaceName source = (PlaceName)req.getPlaceName();
			boolean save = false;
			String	classes_path = null;
			ObjectOutputStream	oos = new ObjectOutputStream(output);
			
			if ( getProperty("STORE.CLASSES").equals("true") ) {
				save = true;
				classes_path = getProperty("DIR.CLASSES");
			}
			
			try {
				AgentClassLoader loader = new AgentClassLoader(source,save,classes_path);
				input.setLoader(loader);
				Agent	anAgent = (Agent)input.readObject();
				startAgent(anAgent, Agent.RESTART);
				oos.writeObject( new Message(Message.MIGRATION_OK, null) );
			} catch (Exception e) {
				oos.writeObject( new Message(Message.MIGRATION_KO, null) );
			} finally {
				oos.flush();
			}
		}
	
		private void handleClassRequest(ClassRequestInfo req) {
			String	className = req.getClassName(), classFile;
			File f;
			InputStream is = null;
			DataOutputStream dos = new DataOutputStream(output);
			int n;
			
			classFile = className.replace('.','/') + ".class";
				
			try {
				f = findFile(classFile);
				is = new BufferedInputStream(new FileInputStream(f));
				n = is.available();
				
				// write class file length
				dos.writeInt(n);
				// now write class file
				int byte_read;
				while ( (byte_read = is.read()) >= 0 ) {
					dos.writeByte(byte_read);
				}
				dos.flush();
					
			} catch (IOException ioe) {
				util.exc(ioe);
			} finally {
				if (is != null ) {
					try {
						is.close();
					} catch (IOException ioe) {}
				}
			}
				
			try {
				dos.close();
			} catch (IOException ioe) {}
		}
		
		private File findFile(String classFileName) throws IOException {
			class_path += ":" + getProperty("DIR.CLASSES");
			StringTokenizer	st = new StringTokenizer(class_path, ":");
			String	file_path, dir;
			File f;
			
			// find directory containing the requested class
			while( st.hasMoreTokens() )	{
			
				dir = st.nextToken();
			
				if ( dir.endsWith(file_separator) ) {
					file_path = dir + classFileName;
				} else {
					file_path = dir + file_separator + classFileName;
				}
				
				f = new File(file_path);
				
				if ( !f.exists() ) {
					continue;
				}
				
				if ( !f.canRead() ) {
					// sorry, no error handling yet ...
				}
				return f;
			}
			return null;
		}
	}
	
	/**
	 *	This class represents a running agent. Using <code>AgentThread</code>
	 *	is possible to manage the execution of an agent as a thread.
	 */

	class AgentThread extends Thread {

		private Agent	agent;
		private int		mode;

		
		// Constructors ------------------------------------------------------
	
		/**
		 *	Default constructor.
		 */
		public AgentThread() {
			super();
		}
	
		/**
		 *	Builds a new <code>AgentThread</code> that execute the given agent.
		 *
		 *	@param	agent	the agent to execute.
		 */
		public AgentThread(Agent agent, int mode) {
			super();
			this.agent = agent;
			this.mode = mode;
		}
	
		// Public Methods ----------------------------------------------------
	
		/**
		 *	Sets the agent to run.
		 *
		 *	@param	agent	the agent to run.
		 */
		public void setAgent(Agent agent) {
			if (this.agent != null) {
				this.agent = agent;
			}
		}

		/**
		 *	Returns the executing agent or null if the agent is not set.
		 *
		 *	@return	the executing agent or null if the agent is not set.
		 */
		public Agent getAgent() {
			return agent;
		}
	
		/**
		 *	Runs the agent.
		 */
		public void run() {
			if (agent != null) {
				try {
					if (mode == Agent.INIT) {
						agent.init();
					} else if (mode == Agent.RESTART) {
						agent.callback();
					} else if (mode == Agent.RESTORE) {
						// ??
					}
				} catch (Exception e) {
				}
			}
		}

		/**
		 *	Returns a string representation of this object.
		 *
		 *	@return	a string representation of this object.
		 */
		public String toString() {
			return (super.toString() + agent.getClass().getName());
		}
	}
	
	class MigrationThread extends Thread {

		private Agent		agent;
		private PlaceName	pname;

		public MigrationThread(Agent agent, PlaceName pname) {
			this.agent = agent;
			this.pname = pname;
		}
		
		public void run() {
		
			Socket			socket = null;
			ObjectInput		in = null;
			ObjectOutput	out = null;
		
			try {
				socket = new Socket(pname.getHost(), pname.getPort());
				out = new ObjectOutputStream(socket.getOutputStream());

				// build and send request
				Message req = new Message( Message.AGENT_ENTER_REQUEST, new AgentEnterInfo(getPlaceName()) );
				out.writeObject(req);
				// now stop the agent
				AgentThread at = (AgentThread)agents.remove(agent.getClass().getName());
				at.suspend();
				// send current agent state
				out.writeObject(agent);
				out.flush();
			
				in = new ObjectInputStream(socket.getInputStream());
				Message reply = (Message)in.readObject();
				if ( reply.getType() == Message.MIGRATION_OK ) {
					// OK! Stop the agent
					mresult.setSuccess();
					at.stop(); // needed ??
				} else {
					// KO :( Now resume the agent
					at.resume();
					agents.put(agent.getClass().getName(), at);
					// and throw an exception to the agent
					mresult.setException(new TransportFailedException());
				}
			} catch (UnknownHostException uhe) {
				util.exc(uhe);
				mresult.setException(new UnknownDestinationException());
			} catch (ClassNotFoundException cnfe) {
			} catch (IOException ioe) {
				// agent can't be moved
				util.exc(ioe);
				mresult.setException(new TransportFailedException());
			} finally {
				// clean up everything
				try {
					in.close();
					out.close();
					socket.close();
				} catch (IOException ioe) {}
			}
		}
	}
			
}
