package macondo.mjada;

import java.util.Vector;
import java.util.Enumeration;
import java.util.StringTokenizer;
import java.io.*;
import jada.JadaItem;

/**
 *	The name of a <code>Space</code> object.
 *
 *	@author		Andrea Giovannini
 *	@version	beta
 */
public class SpaceName implements JadaItem {
	private Vector	name_items;
	private final static String DELIMITER = "/";
	private final static String BACK = "..";
	
	// Constructor ------------------------------------------------------
	
	/**
	 *	Default constructor.
	 */
	public SpaceName() {
		name_items = new Vector();
	}
	
	// Public methods ----------------------------------------------------

	/**
	 *	Joins to the space specified from the given name items.
	 *
	 *	@param	items	the name items.
	 *	@see	Space#join
	 *	@see	#leave()
	 *	@see	#leaveAll()
	 */
	public synchronized void join(String items) {
		StringTokenizer	st = new StringTokenizer(items, DELIMITER);
		String	item;
		
		if ( items.startsWith( DELIMITER ) ) {
			name_items.removeAllElements(); 
		}
		while( st.hasMoreTokens() ) {
			item = st.nextToken();
			if ( item.equals(BACK) ) {
				name_items.removeElementAt( name_items.size()-1 );
			} else {
				name_items.addElement( item );
			}
		}
	}
	
	/**
	 *	Moves to the encompassing tuple space.
	 *
	 *	@see	Space#leave
	 *	@see	#join
	 *	@see	#leaveAll()
	 */
	public synchronized void leave() {
		name_items.removeElementAt( name_items.size()-1 );
	}
	
	/**
	 *	Moves to the root tuple space.
	 *
	 *	@see	Space#leaveAll()
	 *	@see	#join
	 *	@see	#leave()
	 */
	public synchronized void leaveAll() {
		name_items.removeAllElements();
	}

	/**
	 *	Returns a string representation of this <code>SpaceName</code>
	 *	object.
	 *
	 *	@return		a string representation of this <code>SpaceName</code>
	 *				object.
	 */
	public synchronized String toString() {
		String result = "";
		Enumeration enum = name_items.elements();
		
		while( enum.hasMoreElements() ) {
			result += ("/" + (String)enum.nextElement());
		}
		
		return result;
	}

	/**
	 *	Returns the current number of items of this <code>SpaceName</code>.
	 *
	 *	@return		the current number of items in this <code>SpaceName</code>.
	 */
	public int size() {
		return name_items.size();
	}

	/**
	 *	Tests if this name starts with the specified prexif.
	 *
	 *	@param		prefix	the prefix.
	 *	@return		<code>true</code> if the string sequence represented
	 *				by the argument is a prefix of the string sequence
	 *				represented by this name; <code>false</code> otherwise.
	 */
	public synchronized boolean startsWith(SpaceName prefix) {
		Enumeration	my_items = name_items.elements(),
					other_items = prefix.name_items.elements();
		String	item1, item2;
		
		if ( name_items.size() <= prefix.name_items.size() ) {
			return false;
		}
		
		while ( my_items.hasMoreElements() ) {
			item1 = (String)my_items.nextElement();
			item2 = (String)other_items.nextElement();
			if ( !(item1.equals(item2)) ) {
				return false;
			}
		}
		
		return true;
	}
	
	/**
	 *	Returns the last item of this name.
	 *
	 *	@return	the last item of this name.
	 */
	public synchronized String lastItem() {
		return (String)name_items.lastElement();
	}

	/**
	 *	Returns the items of this name as an array of String.
	 *
	 *	@return		the name items as an array of String.
	 */
	public synchronized String[] items() {
		String[]	buffer = new String[name_items.size()];
		
		for(int i=0; i<name_items.size(); i++) {
			buffer[i] = (String)name_items.elementAt(i);
		}
		
		return buffer;
	}
	

	// JadaItem interface methods

	/**
	 *	JadaItem interface method.
	 *	Dumps this object to the specified stream.
	 *
	 *	@param	os	the output stream.
	 *	@exception	ItemNotStreamableException	thrown if the object is not
	 * 					streamable.
	 *	@exception	IOException					thrown if an I/O error
	 *					occurs.
	 */	
	public void dumpItem(OutputStream os) throws ItemNotStreamableException,
			IOException {
		ObjectOutputStream output = new ObjectOutputStream( os );
		output.writeObject( name_items );
	}
	
	/**
	 *	JadaItem interface method.
	 *	Builds a <code>SpaceName</code> object from a stream.
	 *
	 *	@param	is		the input stream.
	 *	@exception	ItemNotStreamableException	thrown if the object is not
	 * 					streamable.
	 *	@exception	IOException					thrown if an I/O error
	 *					occurs.
	 */
	public void buildItem(InputStream is) throws ItemNotStreamableException,
			IOException {
		ObjectInputStream input = new ObjectInputStream(is);
		try {
			name_items = (Vector)input.readObject();
		} catch (ClassNotFoundException cnfe) {
		}
	}
	
	/**
	 *	JadaItem interface method.
	 *	Clone this object from another one.
	 *
	 *	@param	source	the object to clone from.
	 */
	public void cloneItem(Object source) {
		if ( !(source instanceof SpaceName) ) {
			return;
		}
		SpaceName sn = (SpaceName)source;
		name_items = (Vector)sn.name_items.clone();
	}
	
	/**
	 *	JadaItem interface method.
	 *	Tests if this object matches the given one.
	 *
	 *	@param	match	the object to tests.
	 *	@return			<code>true</code> if the two object match,
	 *					false otherwise.
	 */
	public boolean matchesItem(JadaItem match) {
		
		if ( !(match instanceof SpaceName) ) {
			return false;
		}
		
		SpaceName	sn = (SpaceName)match;
		
		if ( name_items.size() != sn.name_items.size() ) {
			return false;
		}

		Enumeration enum1 = name_items.elements(),
					enum2 = sn.name_items.elements();
		String		item1, item2;
		
		while( enum1.hasMoreElements() ) {
			item1 = (String)enum1.nextElement();
			item2 = (String)enum2.nextElement();
			if ( !(item1.equals(item2)) ) {
				return false;
			}
		}
		
		return true;
	}
	
}
