package macondo.mjada.collections;

import macondo.mjada.Tuple;
import macondo.mjada.Space;
import macondo.mjada.SpaceName;
import macondo.mjada.SpaceException;
import jada.ObjectSpace;
import jada.JadaException;

/**
 *	A <code>TupleCollection</code> represents a group of tuples with the
 *	same pattern inside a tuple space. Tuples can be added to a collection
 *	and then read or withdrawn with an iterator in the same order they have
 *	been inserted.
 *
 *	@author		Andrea Giovannini
 *	@version 	beta
 */

public class TupleCollection {

	private Space			space;
	private ObjectSpace		object_space;
	private SpaceName		space_name;
	private Tuple	pattern;
	private Tuple	tokenTuple;

	// Constructor -------------------------------------------------------
	
	/**
	 *	Builds a new <code>TupleCollection</code>.
	 *
	 *	@param	space	the tuple space of the collection.
	 *	@param	token	the token used to enumerate tuples.
	 *	@param	pattern	the pattern of tuples in this collection.
	 */
	public TupleCollection(Space space, Tuple pattern) {
		this.space = space;
		this.object_space = space.getObjectSpace();
		this.space_name = space.getSpaceName();
		this.pattern = pattern;

		Token token = new Token();		
		tokenTuple = new Tuple(space_name, "__token", token.getClass(), pattern);

		Tuple	tmp_tuple = new Tuple(space_name, "__token", token, pattern);
		try {
			if ( object_space.read_nb(tokenTuple) == null ) {
				object_space.out( tmp_tuple );
			}
		} catch (JadaException je) {
		}
	}
	
	// Public Methods ----------------------------------------------------
	
	/**
	 *	Adds a tuple to this collection. The tuple is added if and only
	 *	if it matches the pattern of this collection, otherwise an
	 *	exception is thrown.
	 *
	 *	@param	tuple	the tuple inserted in this collection.
	 *	@exception	IllegalTupleException	thrown if the specified tuple
	 *					doesn't match the pattern of the collection.
	 */
	public void add(Tuple tuple) throws IllegalTupleException {
	
		if ( !(tuple.matchesItem(pattern)) ) {
			throw new IllegalTupleException();
		}
	
		try {
			Tuple currTokenTuple = (Tuple)object_space.in( tokenTuple );
			Token currToken = (Token)currTokenTuple.getItem(2);
			object_space.out( new Tuple( space_name, currToken, tuple ) );
			currToken.advance();
			currTokenTuple.setItem(2, currToken);
			object_space.out( currTokenTuple );
		} catch (JadaException je) {
		}
	}
	
	/**
	 *	Returns a </code>ReadIterator</code> on this collection.
	 *
	 *	@return		a </code>ReadIterator</code> on this collection.
	 *	@see		TupleIterator
	 *	@see		ReadIterator
	 */
	public TupleIterator readIterator() {
		return new ReadIterator(space, pattern);
	}
	
	/**
	 *	Returns a </code>InIterator</code> on this collection.
	 *
	 *	@return		a </code>InIterator</code> on this collection.
	 *	@see		TupleIterator
	 *	@see		InIterator
	 */
	public TupleIterator inIterator() {
		return new InIterator(space, pattern);
	}
	
	/**
	 *	Tests if this collection contains no element.
	 *
	 *	@return		<code>true</code> if this collection contains no element,
	 *				<code>false</code> otherwise.
	 *	@exception	SpaceException	thrown if an error occurs accessing
	 *								the tuple space.
	 */
	public boolean isEmpty() throws SpaceException {
		try {
			if ( object_space.read_nb(new Tuple( space_name, new Token().getClass(), Tuple.formal())) != null) {
				return false;
			}
		} catch (JadaException e) {
			throw new SpaceException();
		}
		return true;
	}
	
	/**
	 *	Tests if this collection contains the specified tuple.
	 *
	 *	@param	tuple	tuple whose presence is to be tested.
	 *	@return			<code>true</code> if the tuple is present in this
	 *					collection, <code>false</code> otherwise. 
	 *	@exception		SpaceException	thrown if an error occurs accessing
	 *					the tuple space.
	 */
	public boolean contains(Tuple tuple) throws SpaceException {
		try {
			if ( object_space.read_nb(new Tuple( space_name, new Token().getClass(), tuple)) != null) {
				return true;
			}
		} catch (JadaException e) {
			throw new SpaceException();
		}
		return false;
	}
	
}